#include <stdint.h>
#include <stdbool.h>
#include <stddef.h>

#include <kernel/paging.h>
#include <kernel/tty.h>
#include <debugging.h>
#include <kernel/heap.h>

struct x86_Page_Directory {
    bool present: 1;
    bool read_write: 1;
    bool user_supervisor: 1;
    bool write_through: 1;
    bool cache_disable: 1;
    bool accesed: 1;
    bool available_2: 1;
    bool page_size: 1;
    uint8_t available_1 : 4;
    uintptr_t address : 20; // Multiply by 0x1000
} __attribute__((packed));

struct x86_Page_Table {
    bool present: 1;
    bool read_write: 1;
    bool user_supervisor: 1;
    bool write_through: 1;
    bool cache_disable: 1;
    bool accesed: 1;
    bool dirty: 1;
    bool page_attribute_table: 1;
    bool global: 1;
    uint8_t available_1 : 3;
    uintptr_t address : 20; // Multiply by 0x1000
} __attribute__((packed));

extern struct x86_Page_Directory boot_page_directory[1024] __attribute__((aligned(4096)));

static struct x86_Page_Table user_code_table __attribute__((aligned(4096)));
static uint8_t user_code[4096] __attribute__((aligned(4096)));

static struct x86_Page_Table user_data_table __attribute__((aligned(4096)));
static uint8_t user_data[4096] __attribute__((aligned(4096)));

static struct x86_Page_Table* get_table(uint16_t index) {
    struct x86_Page_Table* table = (struct x86_Page_Table*)((boot_page_directory[index].address << 12) + 0xC0000000);
    return table;
}

static void reload_pages () {
    asm("\
movl %cr3, %ecx; \
movl %ecx, %cr3; \
    ");
}

static void set_page_directory_entry(uintptr_t address, struct x86_Page_Table* table) {
    boot_page_directory[address >> 22].address = ((uintptr_t)(table) - 0xC0000000) >> 12;
    boot_page_directory[address >> 22].present = true;
    boot_page_directory[address >> 22].read_write = true;
    boot_page_directory[address >> 22].user_supervisor = true;
}


void setup_paging() {
    struct x86_Page_Table* kernel = get_table(768);
    uint16_t i = 0;
    while (kernel[i].present == 0) {
        i++;
    }

    terminal_writestring("Kernel Start Page: ");
    print_hex_bytes(&i, 2);
    while (kernel[i].present != 0) {
        i++;
    }

    terminal_writestring("\nKernel End Page: ");
    print_hex_bytes(&i, 2);
    terminal_putchar('\n');

    set_page_directory_entry(0, &user_code_table);

    user_code_table.address = ((uintptr_t)user_code - 0xC0000000) >> 12;
    user_code_table.present = true;
    user_code_table.read_write = true;
    user_code_table.user_supervisor = true;

    set_page_directory_entry(0x00400000, &user_data_table);

    user_data_table.address = ((uintptr_t)user_data - 0xC0000000) >> 12;
    user_data_table.present = true;
    user_data_table.read_write = true;
    user_data_table.user_supervisor = true;

    user_data[0] = 'B';

    reload_pages();
}
